import flask


class PathNotFoundError(Exception):
    """
    The URL path wasn't recognised
    """

    def __init__(self, path, *args, message=None, **kwargs):
        self.path = path

        if not message:
            message = f"Path {path} not found"

        super().__init__(message, *args, **kwargs)


class RedirectFoundError(Exception):
    """
    If we encounter redirects from Discourse, we need to take action
    """

    def __init__(self, path, target_url, *args, message=None, **kwargs):
        self.path = path
        self.target_url = target_url

        if not message:
            message = f"Path {path} has moved to {target_url}"

        super().__init__(*args, **kwargs)


class MetadataError(Exception):
    """
    If the metadata at the top of the post contains errors.
    These are errors enforced by the Web team for the
    metadata to follow a specific format
    """

    def __init__(self, *args: object) -> None:
        error_message = args[0]
        flask.current_app.extensions["sentry"].captureMessage(
            f"Engage pages metadata error: {error_message}"
        )
        pass


class MarkdownError(Exception):
    """
    If the content of the post contains errors.

    These are errors raised when the author of the post makes a
    Markdown error
    """

    def __init__(self, *args: object) -> None:
        error_message = args[0]
        flask.current_app.extensions["sentry"].captureMessage(
            f"Engage pages markdown error {error_message}"
        )
        pass


class DataExplorerError(Exception):
    pass



# Standard library
import copy
import os
import re
from urllib.parse import urlparse

# Packages
import dateutil.parser
import humanize
from bs4 import BeautifulSoup
from jinja2 import Template

# Local
from canonicalwebteam.discourse.parsers.base_parser import (
    TOPIC_URL_MATCH,
    BaseParser,
)
from canonicalwebteam.discourse.exceptions import (
    PathNotFoundError,
    RedirectFoundError,
)


class DocParser(BaseParser):
    def __init__(
        self,
        api,
        index_topic_id,
        url_prefix,
        tutorials_index_topic_id=None,
        tutorials_url_prefix=None,
    ):
        self.active_topic_id = None
        self.versions = []
        self.navigations = {}
        self.url_map_versions = {}

        # Tutorials
        self.tutorials_url_map = {}
        self.tutorials_index_topic_id = tutorials_index_topic_id
        self.tutorials_url_prefix = tutorials_url_prefix

        return super().__init__(api, index_topic_id, url_prefix)

    def parse(self):
        """
        Get the index topic and split it into:
        - navigation
        - index document content
        - URL map
        - redirects map
        And set those as properties on this object
        """
        self.index_topic = self.api.get_topic(self.index_topic_id)

        raw_index_soup = BeautifulSoup(
            self.index_topic["post_stream"]["posts"][0]["cooked"],
            features="html.parser",
        )

        # Parse navigation and version table (if present)
        self.versions = self._parse_version_table(raw_index_soup)
        self.navigations = self._parse_navigation_versions(raw_index_soup)

        # URL mapping
        self.url_map_versions = self._generate_url_map(self.navigations)
        self.url_map = self._generate_flat_url_map(self.url_map_versions)

        # URL mapping for tutorials
        if self.tutorials_index_topic_id:
            self.tutorials_url_map = self._generate_tutorials_url_map(
                self.tutorials_index_topic_id
            )

        # Parse redirects mappings
        self.redirect_map, redirect_warnings = self._parse_redirect_map(
            raw_index_soup
        )
        self.warnings += redirect_warnings

    def parse_topic(self, topic, docs_version=""):
        """
        Parse a topic object from the Discourse API
        and return document data:
        - title: The title
        - body_html: The HTML content of the initial topic post
                        (with some post-processing)
        - updated: A human-readable date, relative to now
                    (e.g. "3 days ago")
        - forum_link: The link to the original forum post
        """
        self.active_topic_id = topic["id"]

        updated_datetime = dateutil.parser.parse(
            topic["post_stream"]["posts"][0]["updated_at"]
        )

        topic_path = f"/t/{topic['slug']}/{topic['id']}".replace("—", "--")

        topic_soup = BeautifulSoup(
            topic["post_stream"]["posts"][0]["cooked"], features="html.parser"
        )

        # Remove Navigation section from all the index topics
        version_topics = [x["index"] for x in self.versions]

        if topic["id"] in version_topics:
            topic_soup = self._get_preamble(
                topic_soup,
                break_on_title="Navigation",
            )

        # Set navigation for the current version
        self.navigation = self._generate_navigation(
            self.navigations, docs_version
        )

        soup = self._process_topic_soup(topic_soup)

        if self.tutorials_index_topic_id:
            self._parse_tutorials(topic_soup)

        self._replace_lightbox(soup)
        sections = self._get_sections(soup)
        headings_map = self._generate_headings_map(soup)
        metadata = self._parse_docs_metadata(soup)

        return {
            "title": topic["title"],
            "body_html": str(soup),
            "sections": sections,
            "headings_map": headings_map,
            "updated": humanize.naturaltime(
                updated_datetime.replace(tzinfo=None)
            ),
            "topic_id": topic["id"],
            "topic_path": topic_path,
            "metadata": metadata,
        }

    def resolve_path(self, relative_path):
        """
        Given a path to a Discourse topic, and a mapping of
        URLs to IDs and IDs to URLs, resolve the path to a topic ID

        A PathNotFoundError will be raised if the path is not recognised.

        A RedirectFoundError will be raised if the topic should be
        accessed at a different URL path.
        """
        version = ""
        version_paths = [x["path"] for x in self.versions]
        version_path = relative_path.lstrip("/").split("/")[0]

        if version_path in version_paths:
            version = version_path

        full_path = os.path.join(self.url_prefix, relative_path.lstrip("/"))

        if full_path in self.redirect_map:
            raise RedirectFoundError(
                full_path, target_url=self.redirect_map[full_path]
            )
        elif full_path in self.url_map_versions[version]:
            topic_id = self.url_map_versions[version][full_path]
        else:
            topic_id = self._get_url_topic_id(relative_path)

            if topic_id in self.url_map_versions[version]:
                raise RedirectFoundError(
                    full_path,
                    target_url=self.url_map_versions[version][topic_id],
                )
            else:
                raise PathNotFoundError(full_path)

        return topic_id, version

    def resolve_path_all_versions(self, relative_path, current_version):
        """
        Given a path to a pretty URL try to obtain all the
        paths using this URL for all the different versions.

        If the path doesn't exist the output will be the default
        version index path

        eg: /machines
        {"v1": "/v1/machines/", "": "/machines/"}
        """
        result = {}

        if relative_path.startswith(f"/{current_version}/"):
            chars_to_remove = len(current_version) + 1
            relative_path = relative_path[chars_to_remove:]

        for version in self.url_map_versions.keys():
            if version:
                version_relative_path = f"/{version}{relative_path}"
            else:
                version_relative_path = relative_path

            version_relative_path = f"{self.url_prefix}{version_relative_path}"

            if version_relative_path in self.url_map_versions[version]:
                result[version] = version_relative_path
            else:
                result[version] = f"{self.url_prefix}/{version}"

        return result

    def _generate_url_map(self, navigations):
        """
        Given all the navigation versions defined
        this method will iterate over them and call
        _parse_topic_url_map to process each of them.
        """

        url_map = {}

        for version_path, navigation in navigations.items():
            url_map[version_path] = {}

            if version_path:
                url_prefix = f"{self.url_prefix}/{version_path}"
            else:
                url_prefix = self.url_prefix

            for item in navigation["nav_items"]:
                pretty_path = item["path"]
                topic_url = item["navlink_href"]

                # URL has a path but is not linked to a topic
                if pretty_path and not topic_url:
                    self.warnings.append(
                        f"Missing topic link for {pretty_path}"
                    )
                    continue

                # There is a link to a topic without path
                if topic_url and not pretty_path:
                    # It's fine to not specify a path for an external link
                    if not self._match_url_with_topic(topic_url):
                        continue

                    # It's fine to not specify a path for the main topic
                    if self.index_topic != self._get_url_topic_id(topic_url):
                        self.warnings.append(
                            f"Missing topic path for {topic_url}"
                        )
                        continue

                # No need to map them when missing
                if not topic_url or not pretty_path:
                    continue

                topic_path = urlparse(topic_url).path
                topic_match = TOPIC_URL_MATCH.match(topic_path)

                if not pretty_path.startswith("/"):
                    pretty_path = "/" + pretty_path
                if not pretty_path.startswith(url_prefix):
                    pretty_path = url_prefix + pretty_path

                if not topic_match or not pretty_path.startswith(url_prefix):
                    self.warnings.append(
                        f"Could not parse URL map item {item}"
                    )
                    continue

                topic_id = int(topic_match.groupdict()["topic_id"])

                url_map[version_path][pretty_path] = topic_id

            # Add the reverse mappings as well, for efficiency
            ids_to_paths = dict(
                [reversed(pair) for pair in url_map[version_path].items()]
            )
            url_map[version_path].update(ids_to_paths)

            # Add the homepage path
            home_path = url_prefix

            if home_path != "/" and home_path.endswith("/"):
                home_path = home_path.rstrip("/")

            url_map[version_path][home_path] = navigation["index"]
            url_map[version_path][navigation["index"]] = home_path

        return url_map

    def _generate_flat_url_map(self, url_map_versions):
        """
        This method generates a flatten URL map for
        compatibility with the other parsers, so things
        like sitemap generation work as expected.
        """
        url_map = {}

        for version, version_url_map in url_map_versions.items():
            url_map.update(version_url_map)

        return url_map

    def _parse_navigation_versions(self, main_index_soup):
        """
        Given the HTML soup of an index topic
        extract the navigation table from the "Navigation"
        section.

        The navigation section should contain a table of
        "Level", "Path" and "Navlink" mappings
        (extra markup around this table doesn't matter)

        # Navigation

        [details=Navigation]
        | Level | Path | Navlink |
        | -- | -- | -- |
        | 1 | | Getting started |
        | 1 | install | [Install](/t/install-the-example-charm/100000) |
        | 2 | install/gke | [GKE](/t/install-the-example-charm-on-gke/100002) |
        | 2 | install/aks | [AKS](/t/install-the-example-charm-on-gke/100002) |
        | 2 | install/eks | [EKS](/t/install-the-example-charm-on-gke/100002) |
        [/details]

        Optionally, it can contain a version table linking to
        previous docs topics:
        [details=Documentation versions]
        | Path | Version |
        |--|--|
        |  | 2.x, 3.x and dev |
        | v1 | [1.x and older](/t/1x-doc-nav) |
        [/details]
        """
        navigations = {}

        for version in self.versions:
            version["nav_items"] = self._parse_navigation_table(
                version["index"], main_index_soup
            )
            navigations[version["path"]] = version

        return navigations

    def _parse_navigation_table(self, topic_id, main_index_soup):
        """
        Given the HTML soup of an index topic
        extract the navigation table from the "Navigation" section.

        The URLs section should contain a table of
        "Path" to "Location" mappings
        (extra markup around this table doesn't matter)

        # Navigation

        [details=Navigation]
        | Level | Path | Navlink |
        | -- | -- | -- |
        | 1 | | Getting started |
        | 1 | install | [Install](/t/install-the-example-charm/100000) |
        | 2 | install/gke | [GKE](/t/install-the-example-charm-on-gke/100002) |
        | 2 | install/aks | [AKS](/t/install-the-example-charm-on-gke/100002) |
        | 2 | install/eks | [EKS](/t/install-the-example-charm-on-gke/100002) |
        [/details]
        """
        nav_items = []

        if topic_id == self.index_topic_id:
            index_soup = main_index_soup
        else:
            index_topic = self.api.get_topic(topic_id)
            index_soup = BeautifulSoup(
                index_topic["post_stream"]["posts"][0]["cooked"],
                features="html.parser",
            )

        navigation_soup = self._get_section(index_soup, "Navigation")

        if navigation_soup:
            navigation_table = []
            tables = navigation_soup.findAll("table")

            for table in tables:
                if table.select("tr:has(> th:-soup-contains('Navlink'))"):
                    navigation_table = table.select("tr:has(td)")

            for row in navigation_table:
                item = {}
                level = row.select_one("td:first-child").text
                hidden = False

                # Empty levels are possible to allow URLs mapping
                if len(level) == 0:
                    hidden = True
                    level = "0"

                if not level.isnumeric() or int(level) < 0:
                    self.warnings.append(f"Invalid level used: {level}")
                    continue

                path = row.select_one("td:nth-of-type(2)").text.replace(
                    "–", "--"
                )
                navlink_cell = row.select_one("td:last-child")

                navlink_href = navlink_cell.find("a", href=True)
                if navlink_href:
                    navlink_href = navlink_href.get("href", "").replace(
                        "–", "--"
                    )

                navlink_text = navlink_cell.text

                parsed_path = urlparse(path)
                parsed_href = urlparse(navlink_href)
                item["hidden"] = hidden
                item["level"] = int(level) if level else None
                item["path"] = parsed_path.path
                item["navlink_href"] = navlink_href
                item["navlink_fragment"] = parsed_href.fragment
                item["navlink_text"] = navlink_text if not hidden else ""
                item["is_active"] = False
                item["has_active_child"] = False
                item["children"] = []

                nav_items.append(item)

        return nav_items

    def _parse_version_table(self, main_index_soup):
        """
        Given the HTML soup of an index topic
        extract the version table from the "Navigation"
        section.

        The navigation section should contain a table of
        "Path" and "Version" mappings
        (extra markup around this table doesn't matter)

        # Navigation

        [details=Documentation versions]
        | Path | Version |
        |--|--|
        |  | 2.x, 3.x and dev |
        | v1 | [1.x and older](/t/1x-doc-nav) |
        [/details]
        """
        navigation_soup = self._get_section(main_index_soup, "Navigation")

        # Default version of docs
        versions = [
            {
                "index": int(self.index_topic_id),
                "path": "",
                "version": "latest",
                "nav_items": [],
            }
        ]

        if not navigation_soup:
            return versions

        # Get and identify the version table
        version_table = None
        tables = navigation_soup.findAll("table")

        for table in tables:
            first_row = table.tr
            if not first_row:
                continue
            headers = first_row("th")
            if not headers:
                continue
            if headers[-1].string == "Version":
                version_table = table("tr")[1:]
        # Parse version table or return a default one if it's missing
        if version_table:
            versions = []

            for row in version_table:
                topic_id = None
                path = row.select_one("td:first-child").text
                version_cell = row.select_one("td:last-child")

                version_href = version_cell.find("a", href=True)
                if version_href:
                    version_href = version_href.get("href")
                    topic_id = self._get_url_topic_id(version_href)

                if not topic_id or not path:
                    topic_id = self.index_topic_id

                versions.append(
                    {
                        "index": int(topic_id),
                        "path": path,
                        "version": version_cell.text,
                        "nav_items": [],
                    }
                )

        return versions

    def _process_nav_levels(self, nav_items):
        """
        Given a list of nav_items, it will generate a tree structure
        """
        root = {}
        root["children"] = []

        if nav_items:
            first_level = nav_items[0]["level"]

            # We need a level 0 parent to group them
            if first_level != 0:
                nav_items.insert(
                    0,
                    {
                        "level": 0,
                        "path": "",
                        "navlink_href": None,
                        "navlink_text": None,
                        "children": [],
                    },
                )

            for node in nav_items:
                last = root
                for _ in range(node["level"]):
                    if last["children"]:
                        last = last["children"][-1]
                last["children"].append(node)

        return root["children"]

    def _has_active_nav_child(self, item):
        if item["children"]:
            for child in item["children"]:
                if child["is_active"] or self._has_active_nav_child(child):
                    item["has_active_child"] = True
                    return True

        item["has_active_child"] = False
        return False

    def _generate_navigation(self, navigations, version_path):
        # we mutate the navigations[version_path] dictionary and so to
        # avoid retaining state between document views, we need to
        # take a (deep)copy.
        navigation = copy.deepcopy(navigations[version_path])

        # Replace links with url_map
        for item in navigation["nav_items"]:
            if item["navlink_href"] and self._match_url_with_topic(
                item["navlink_href"]
            ):
                topic_id = self._get_url_topic_id(item["navlink_href"])
                if topic_id in self.url_map_versions[version_path]:
                    href = self.url_map_versions[version_path][topic_id]
                    fragment = item["navlink_fragment"]
                    if fragment:
                        item["navlink_href"] = f"{href}#{fragment}"
                    else:
                        item["navlink_href"] = href

                # Check if given item should be marked as active
                if topic_id == self.active_topic_id:
                    item["is_active"] = True

        # Generate tree structure with levels
        navigation["nav_items"] = self._process_nav_levels(
            navigation["nav_items"]
        )

        # Check for any active children
        for item in navigation["nav_items"]:
            self._has_active_nav_child(item)

        return navigation

    def _parse_tutorials(self, soup):
        """
        Get a list of tutorials topic IDs from all the
        tutorial tables in a topic

        Example of a tutorial table:
        | Tutorials |
        | -- |
        | https://discourse.charmhub.io/t/add-docs-to-your-charm-page/3784 |
        """
        tutorial_tables = []

        tables = soup.select("table:has(th:-soup-contains('Tutorials'))")
        if soup.select("table:has(th:nth-child(2))"):
            return

        for table in tables:
            table_rows = table.select("tr:has(td)")

            if table_rows:
                tutorial_set = {"soup_table": table, "topics": []}

                # Get all tutorial topics in this table
                for row in table_rows:
                    navlink_href = row.find("a", href=True)

                    if navlink_href:
                        navlink_href = navlink_href.get("href")

                        try:
                            topic_id = self._get_url_topic_id(navlink_href)
                        except PathNotFoundError:
                            self.warnings.append("Invalid tutorial URL")
                            continue

                        tutorial_set["topics"].append(topic_id)

                tutorial_tables.append(tutorial_set)

        if tutorial_tables:
            # Get tutorials metadata from Data Explorer API
            tutorial_data = self._parse_tutorials_metadata(tutorial_tables)

            # Remplace tables with cards
            self._replace_tutorials(tutorial_tables, tutorial_data)

    def _parse_tutorials_metadata(self, tutorial_tables):
        """
        Get multiple tutorials from one API call and
        parse their metadata table

        Example of metadata table:
        | — | ----------------------- |
        | Summary | Learn how to deploy |
        | Categories | cloud |
        | Difficulty | 2 |
        | Author | John |
        """
        if not self.api.get_topics_query_id:
            self.warnings.append(
                "Tutorials found but Data Explorer query is not set"
            )

        # Topics that we need from the API
        topics = []

        for table in tutorial_tables:
            topics += table["topics"]

        response = self.api.get_topics(topics)
        tutorial_data = {}

        for topic in response:
            topic_soup = BeautifulSoup(
                topic[3],
                features="html.parser",
            )

            # Get table with tutorial metadata
            rows = topic_soup.select("table:first-child tr:has(td)")

            if not rows:
                self.warnings.append(
                    f"Invalid metadata table for tutorial topic {topic[0]}"
                )
                continue

            link = self.tutorials_url_map.get(
                topic[0], f"{self.api.base_url}/t/{topic[2]}/{topic[0]}"
            )

            metadata = {"title": topic[1], "link": link}
            for row in rows:
                key = row.select_one("td:first-child").text.lower()
                value = row.select_one("td:last-child").text
                metadata[key] = value

            tutorial_data[topic[0]] = metadata

        return tutorial_data

    def _generate_tutorials_url_map(self, index_topic_id):
        index_topic = self.api.get_topic(index_topic_id)
        raw_index_soup = BeautifulSoup(
            index_topic["post_stream"]["posts"][0]["cooked"],
            features="html.parser",
        )

        url_map, url_warnings = self._parse_url_map(
            raw_index_soup, self.tutorials_url_prefix, index_topic_id, "URLs"
        )

        self.warnings.extend(url_warnings)

        return url_map

    def _replace_tutorials(self, tutorial_tables, tutorial_data):
        """
        Replace tutorial tables to cards
        """
        card_template = Template(
            (
                '<div class="row">'
                "{% for tutorial in tutorials %}"
                '<div class="col-4 col-medium-3 p-card">'
                '<div class="p-card__content">'
                '<h3 class="p-card__title p-heading--four">'
                '<a class="inline-onebox" href="{{tutorial.link}}">'
                "{{ tutorial.title }}</a></h3>"
                "<p>{{ tutorial.summary }}</p>"
                "</div>"
                "</div>"
                "{% endfor %}"
                "</div>"
            )
        )

        for table in tutorial_tables:
            table_cards = [tutorial_data[topic] for topic in table["topics"]]

            card = card_template.render(
                tutorials=table_cards,
            )
            table["soup_table"].replace_with(
                BeautifulSoup(card, features="html.parser")
            )

    def _generate_headings_map(self, soup):
        headings_map = []
        current_list = headings_map
        previous_tag = None

        headings = soup.find_all(["h2", "h3"])

        for heading in headings:
            current_tag = re.findall(r"\d", heading.name)[0]
            if previous_tag and previous_tag < current_tag:
                current_list = []
            elif previous_tag and previous_tag > current_tag:
                current_list = headings_map

            if heading.a and heading.a.has_attr("name"):
                current_list.append(
                    (
                        {
                            "heading_level": current_tag,
                            "heading_text": re.sub("\n", "", heading.text),
                            "heading_slug": heading.a["name"],
                        }
                    )
                )

                if previous_tag and previous_tag < current_tag:
                    headings_map[-1]["children"] = current_list

                previous_tag = current_tag

        return headings_map

    def _parse_docs_metadata(self, index_soup, section_name="Metadata"):
        """
        Given the HTML soup of an index topic
        extract the metadata from the name designated
        by section_name. Currently this is set to the default "Metadata

        This section_name section should contain a table
        e.g.:

        <h1>Metadata</h1>
        <table>
        <tr><th>Column 1</th><th>Column 2</th></tr>
        <tr>
            <td>data 1</td>
            <td>data 2</td>
        </tr>
        <tr>
            <td>data 3</td>
            <td>data 4</td>
        </tr>
        </table>

        This will typically be generated in Discourse from Markdown similar to
        the following:

        # Metadata

        | Key | Value |
        | -- | -- |
        | data 1 | data 2 |
        | data 3 | data 4 |

        The function will return a dictionaries of this format:
        {"data 1": "data 2", "data 3": "data 4"},

        """

        def check_bool(value):
            if value == "true":
                return True
            elif value == "false":
                return False
            else:
                return value

        heading = index_soup.find(re.compile("^h[1-6]$"), string=section_name)

        if not heading:
            return None

        metadata_soup = heading.findNext("div", {"class": "md-table"})

        metadata_object = {}
        if metadata_soup:
            for row in metadata_soup.select("tr:has(td)"):
                key = row.select_one("td:nth-of-type(1)").text.lower()
                value = row.select_one("td:nth-of-type(2)").text.lower()
                metadata_object[key] = check_bool(value)

            heading.decompose()
            metadata_soup.decompose()

        return metadata_object



import flask
import html
from requests.exceptions import HTTPError

from canonicalwebteam.discourse.exceptions import (
    PathNotFoundError,
    RedirectFoundError,
    MetadataError,
    MarkdownError,
)

from canonicalwebteam.discourse.parsers.base_parser import BaseParser
import dateutil.parser
from bs4 import BeautifulSoup, element
from datetime import datetime


class Discourse:
    def __init__(
        self,
        parser,
        document_template,
        url_prefix,
        blueprint_name,
    ):
        self.blueprint = flask.Blueprint(blueprint_name, __name__)
        self.url_prefix = url_prefix
        self.parser = parser

        @self.blueprint.route("/sitemap.txt")
        def sitemap_view():
            """
            Show a list of all URLs in the URL map
            """

            self.parser.parse()

            urls = []

            for key, value in self.parser.url_map.items():
                if type(key) is str:
                    urls.append(flask.request.host_url.strip("/") + key)

            return (
                "\n".join(urls),
                {"Content-Type": "text/plain; charset=utf-8"},
            )

        @self.blueprint.route("/sitemap.xml")
        def sitemap_xml():
            """
            Show a list of all URLs in the URL map
            """

            self.parser.parse()
            pages = []

            for key, value in self.parser.url_map.items():
                if type(key) is str:
                    try:
                        response = parser.api.get_topic(str(value))
                        last_updated = response["post_stream"]["posts"][0][
                            "updated_at"
                        ]
                    except Exception:
                        last_updated = None

                    pages.append(
                        {
                            "url": html.escape(
                                flask.request.host_url.strip("/") + key
                            ),
                            "last_updated": last_updated,
                        }
                    )

            from jinja2 import Template

            tm = Template(
                '<?xml version="1.0" encoding="utf-8"?>'
                '<urlset xmlns="http://www.sitemaps.org/schemas/sitemap/0.9" '
                'xmlns:xhtml="http://www.w3.org/1999/xhtml">'
                "{% for page in pages %}"
                "<url>"
                "<loc>{{ page['url'] }}</loc>"
                "<changefreq>weekly</changefreq>"
                "<lastmod>{{ page['last_updated'] }}</lastmod>"
                "</url>"
                "{% endfor %}"
                "</urlset>"
            )
            xml_sitemap = tm.render(pages=pages)

            response = flask.make_response(xml_sitemap)
            response.headers["Content-Type"] = "application/xml"
            response.headers["Cache-Control"] = "public, max-age=43200"

            return response

    def init_app(self, app):
        """
        Attach the discourse docs blueprint to the application
        at the specified `url_prefix`
        """

        app.register_blueprint(self.blueprint, url_prefix=self.url_prefix)

    def _set_parser_warnings(self, response):
        """
        Append parser warnings to the reponse headers

        :param response: A flask response object
        """

        # To not make the response too big
        # we show only the last ten warnings
        warnings = self.parser.warnings[-10:]

        for message in warnings:
            flask.current_app.logger.warning(message)
            response.headers.add(
                "discourse-warning",
                message,
            )

        # Reset parser warnings
        self.parser.warnings = []

        return response


class Docs(Discourse):
    """
    A Flask extension object to create a Blueprint
    to serve documentation pages, pulling the documentation content
    from Discourse.

    :param api: A DiscourseAPI for retrieving Discourse topics
    :param index_topic_id: ID of a forum topic containing nav & URL map
    :param url_prefix: URL prefix for hosting under (Default: /docs)
    :param document_template: Path to a template for docs pages
                              (Default: docs/document.html)
    """

    def __init__(
        self,
        parser,
        document_template="docs/document.html",
        url_prefix="/docs",
        blueprint_name="docs",
    ):
        super().__init__(parser, document_template, url_prefix, blueprint_name)

        @self.blueprint.route("/")
        @self.blueprint.route("/<path:path>")
        def document_view(path=""):
            """
            A Flask view function to serve
            topics pulled from Discourse as documentation pages.
            """
            docs_version = ""
            path = "/" + path
            self.parser.parse()

            if path == "/":
                document = self.parser.parse_topic(self.parser.index_topic)
            else:
                try:
                    topic_id, docs_version = self.parser.resolve_path(path)
                except RedirectFoundError as redirect:
                    return flask.redirect(redirect.target_url)
                except PathNotFoundError:
                    return flask.abort(404)

                if topic_id == self.parser.index_topic_id:
                    return flask.redirect(self.url_prefix)

                try:
                    topic = self.parser.api.get_topic(topic_id)
                except HTTPError as http_error:
                    return flask.abort(http_error.response.status_code)

                document = self.parser.parse_topic(topic, docs_version)

                if (
                    topic_id not in self.parser.url_map_versions[docs_version]
                    and document["topic_path"] != path
                ):
                    return flask.redirect(document["topic_path"])

            version_paths = self.parser.resolve_path_all_versions(
                path, docs_version
            )

            response = flask.make_response(
                flask.render_template(
                    document_template,
                    document=document,
                    versions=self.parser.versions,
                    navigation=self.parser.navigation,
                    forum_url=self.parser.api.base_url,
                    metadata=self.parser.metadata,
                    docs_version=docs_version,
                    version_paths=version_paths,
                )
            )

            return self._set_parser_warnings(response)


class Tutorials(Discourse):
    """
    A Flask extension object to create a Blueprint
    to serve documentation pages, pulling the documentation content
    from Discourse.

    :param api: A DiscourseAPI for retrieving Discourse topics
    :param index_topic_id: ID of a forum topic containing nav & URL map
    :param url_prefix: URL prefix for hosting under (Default: /docs)
    :param document_template: Path to a template for docs pages
                              (Default: docs/document.html)
    """

    def __init__(
        self,
        parser,
        document_template="tutorials/tutorial.html",
        url_prefix="/tutorials",
        blueprint_name="tutorials",
    ):
        super().__init__(parser, document_template, url_prefix, blueprint_name)

        @self.blueprint.route("/")
        @self.blueprint.route("/<path:path>")
        def document_view(path=""):
            """
            A Flask view function to serve
            topics pulled from Discourse as documentation pages.
            """

            path = "/" + path
            self.parser.parse()

            if path == "/":
                document = self.parser.parse_topic(self.parser.index_topic)
            else:
                try:
                    topic_id = self.parser.resolve_path(path)
                except RedirectFoundError as redirect:
                    return flask.redirect(redirect.target_url)
                except PathNotFoundError:
                    return flask.abort(404)

                if topic_id == self.parser.index_topic_id:
                    return flask.redirect(self.url_prefix)

                try:
                    topic = self.parser.api.get_topic(topic_id)
                except HTTPError as http_error:
                    return flask.abort(http_error.response.status_code)

                document = self.parser.parse_topic(topic)

                if (
                    topic_id not in self.parser.url_map
                    and document["topic_path"] != path
                ):
                    return flask.redirect(document["topic_path"])

            response = flask.make_response(
                flask.render_template(
                    document_template,
                    document=document,
                    forum_url=self.parser.api.base_url,
                    metadata=self.parser.metadata,
                    tutorials=self.parser.tutorials,
                )
            )

            return self._set_parser_warnings(response)


class EngagePages(BaseParser):
    """
    Parsing and rendering of engage pages (parse_engage_pages) and
    takeovers (parse_takeovers)

    :param api: A DiscourseAPI for retrieving Discourse topics
    :param category_id: ID of a forum topic containing nav & URL map
    :param url_prefix: URL prefix on project (Default: /engage)
    :param page_type: ["engage-pages", "takeovers"]. This separation
           is necessary because metadata table is different for each.
    :param skip_posts: Skip given posts from throwing errors
    """

    def __init__(
        self,
        api,
        category_id,
        page_type,
        exclude_topics=[],
        additional_metadata_validation=[],
    ):
        self.api = api
        self.category_id = category_id
        self.page_type = page_type
        self.exclude_topics = exclude_topics
        self.additional_metadata_validation = additional_metadata_validation
        pass

    def get_index(self):
        """
        Get the index topic and split it into:
        - index document content
        - URL map
        And set those as properties on this object
        """
        list_topics = self.api.engage_pages_by_category(self.category_id)
        topics = []
        for topic in list_topics:
            if topic[6] not in self.exclude_topics:
                try:
                    topics_index = self.parse_topics(topic)
                    topics.append(topics_index)
                except MetadataError:
                    continue

        return topics

    def get_engage_page(self, path):
        """
        Get single engage page using data-explorer
        """
        single_topic = self.api.get_engage_pages_by_param(
            self.category_id, "path", path
        )
        try:
            single_topic[0]
        except KeyError:
            return None
        # No metadata found if single_topic = []
        except IndexError:
            return None

        metadata = self.parse_topics(single_topic[0])

        return metadata

    def parse_active_takeovers(self):
        active_takeovers_topics = self.api.get_engage_pages_by_param(
            self.category_id, "active", "true"
        )

        topics = []
        for topic in active_takeovers_topics:
            if topic[6] not in self.exclude_topics:
                try:
                    topics_index = self.parse_topics(topic)
                    topics.append(topics_index)
                except MetadataError:
                    continue

        return topics

    def process_ep_topic_soup(self, soup):
        """
        Given topic HTML soup, apply post-process steps
        """

        soup = self._replace_notifications(soup)
        soup = self._replace_notes_to_editors(soup)
        soup = self._replace_polls(soup)

        return soup

    def parse_topics(self, topic):
        """
        Parse topics in the given category and extract metadata
        to create an index

        Args:
        - topic: "cooked" post content

        returns:
        - topics: list
        """

        # Construct path using slug and id
        topic_path = f"{self.api.base_url}/t/{topic[7]}/{topic[6]}"

        updated_datetime = dateutil.parser.parse(topic[5])

        created_datetime = dateutil.parser.parse(topic[4])

        topic_soup = BeautifulSoup(topic[0], features="html.parser")

        metadata = {}

        # Does metadata table exist?
        try:
            topic_soup.contents[0]("th")[0].text
        except IndexError:
            error_message = f"{topic_path} metadata not found"
            raise MetadataError(error_message)

        if self.page_type == "takeovers":
            # Parse engage pages
            for row in topic_soup.contents[0]("tr"):
                # This condition skips the th key and value headers
                if len(row("td")) > 0:
                    try:
                        key = row("td")[0].contents[0]
                        value = row("td")[1].contents
                        # Allows metadata values to be empty
                        if len(value) == 0:
                            value = ""
                        elif len(value) > 0 and isinstance(
                            value[0], element.Tag
                        ):
                            # Remove <a> links
                            value = value[0].string

                        else:
                            value = value[0]
                        metadata[key] = value
                    except Exception as error:
                        # Catch all metadata errors
                        error_message = (
                            f"{self.page_type} Metadata table contains errors:"
                            f" {error} for {topic_path}"
                        )
                        raise MetadataError(error_message)

            # Further metadata checks
            try:
                self.takeovers_healthcheck(metadata, topic[6])
            except MarkdownError:
                pass

            soup = self.process_ep_topic_soup(topic_soup)
            self._replace_lightbox(soup)

            first_table = soup.select_one("table:nth-of-type(1)")
            if (
                first_table.findAll("th")[0].getText() == "Key"
                and first_table.findAll("th")[1].getText() == "Value"
            ):
                first_table.decompose()

            # Combined metadata old index topic + topic metadata
            metadata.update(
                {
                    "updated": updated_datetime,
                    "created": created_datetime,
                    "topic_id": topic[6],
                    "topic_path": topic_path,
                }
            )
        else:
            # Parse engage pages
            for row in topic_soup.contents[0]("tr"):
                # This condition skips the th key and value headers
                if len(row("td")) > 0:
                    try:
                        key = row("td")[0].contents[0]
                        value = row("td")[1].contents
                        if len(value) == 0:
                            value = ""
                        elif len(value) > 0 and isinstance(
                            value[0], element.Tag
                        ):
                            # Remove <a> links
                            value = value[0].string
                        else:
                            value = value[0]
                        metadata[key] = value
                    except Exception as error:
                        error_message = (
                            "Metadata table contains errors:"
                            f" {error} for {topic_path}"
                        )
                        raise MetadataError(error_message)

            # Further metadata checks
            try:
                self.engage_pages_healthcheck(metadata, topic[6])
            except MarkdownError:
                pass

            soup = self.process_ep_topic_soup(topic_soup)
            self._replace_lightbox(soup)

            first_table = soup.select_one("table:nth-of-type(1)")
            if (
                first_table.findAll("th")[0].getText() == "Key"
                and first_table.findAll("th")[1].getText() == "Value"
            ):
                first_table.decompose()

            # Combined metadata old index topic + topic metadata
            metadata.update(
                {
                    "body_html": str(soup),
                    "updated": updated_datetime,
                    "created": created_datetime,
                    "topic_id": topic[6],
                    "topic_path": topic_path,
                }
            )

        return metadata

    def get_topic(self, topic_id):
        """
        Receives a single topic_id and
        @return the content of the topic
        """
        index_topic = self.api.get_topic(topic_id)
        return self.parse_topic(index_topic)

    def _parse_related(self, tags):
        """
        Filter index topics by tag
        This provides a list of "Related engage pages"
        """
        index_list = [item for item in self.metadata if item["tags"] in tags]
        return index_list

    def engage_pages_healthcheck(self, metadata, topic_id):
        """
        Check engage pages metadata (key/value table)
        for errors
        """
        errors = []

        if "path" not in metadata:
            error = (
                "Missing path on "
                f"https://discourse.ubuntu.com/t/{topic_id}."
                f" This engage page will not show in {self.page_type}"
            )
            errors.append(error)

        if "topic_name" not in metadata:
            error = (
                "Missing topic_name on "
                f"https://discourse.ubuntu.com/t/{topic_id}."
                " Default discourse title will be used"
            )
            errors.append(error)

        if "type" not in metadata:
            error = (
                "Missing type on "
                f"https://discourse.ubuntu.com/t/{topic_id}."
                " Provide a type for this engage page (whitepaper, "
                "webinar, blog, event etc)"
            )
            errors.append(error)

        if "active" not in metadata:
            error = (
                "Missing active on "
                f"https://discourse.ubuntu.com/t/{topic_id}."
                " Provide the active parameter in the metadata (true, false)"
            )
            errors.append(error)

        if "publish_date" in metadata:
            try:
                datetime(metadata["publish_date"])
            except TypeError:
                error = (
                    "publish_date must be a date"
                    " with the following format: yyyy-mm-dd"
                )

        for key in self.additional_metadata_validation:
            if key not in metadata:
                error = (
                    f"Missing {key} on "
                    f"https://discourse.ubuntu.com/t/{topic_id}. "
                    "This parameter is required to render takeovers"
                )
                errors.append(error)

        if len(errors) > 0:
            raise MarkdownError((", ").join(errors))

        pass

    def takeovers_healthcheck(self, metadata, topic_id, title=None):
        """
        Check takeovers metadata (key/value table)
        for errors
        """
        errors = []

        if "title" not in metadata:
            error = (
                "Missing title on "
                f"https://discourse.ubuntu.com/t/{topic_id}. "
                "This takeover will not be displayed"
            )
            errors.append(error)

        if "active" not in metadata:
            error = (
                "Missing active on "
                f"https://discourse.ubuntu.com/t/{topic_id}. "
                "This takeover will not be displayed"
            )
            errors.append(error)

        for key in self.additional_metadata_validation:
            if key not in metadata:
                error = (
                    f"Missing {key} on "
                    f"https://discourse.ubuntu.com/t/{topic_id}. "
                    "This parameter is required to render takeovers"
                )
                errors.append(error)

        if len(errors) > 0:
            raise MarkdownError((", ").join(errors))

        pass



import json
import unittest
from unittest.mock import MagicMock
import warnings

from bs4 import BeautifulSoup
import httpretty
import requests

from canonicalwebteam.discourse.models import DiscourseAPI
from canonicalwebteam.discourse.parsers.base_parser import BaseParser
from canonicalwebteam.discourse.parsers.docs import DocParser

EXAMPLE_CONTENT = """
<p>Some homepage content</p>
<h2>Navigation</h2>

<details>
  <summary>Navigation items</summary>
  <div class="md-table">
    <table>
      <thead>
        <tr>
          <th>Level</th>
          <th>Path</th>
          <th>Navlink</th>
        </tr>
      </thead>
      <tbody>
        <tr>
          <td>0</td>
          <td>/a</td>
          <td><a href="/t/page-a/10">Page A</a></td>
        </tr>
        <tr>
          <td>1</td>
          <td>/page-z</td>
          <td><a href="/t/page-z/26">Page Z</a></td>
        </tr>
      </tbody>
    </table>
  </div>
</details>

<h2>Redirects</h2>
<details>
  <summary>Mapping table</summary>
  <div class="md-table">
    <table>
      <thead>
        <tr>
          <th>PATH</th>
          <th>LOCATION</th>
        </tr>
      </thead>
      <tbody>
        <tr>
          <td>/redir-a</td>
          <td>/a</td>
        </tr>
        <tr>
          <td>/example/page</td>
          <td>https://example.com/page</td>
        </tr>
      </tbody>
    </table>
  </div>
</details>
"""


class TestBaseParser(unittest.TestCase):
    def test_parser_username_link(self):
        discourse_api = DiscourseAPI("https://base.url", session=MagicMock())

        parser = BaseParser(
            api=discourse_api,
            index_topic_id=1,
            url_prefix="/",
        )

        parsed_topic = parser.parse_topic(
            {
                "id": 1,
                "category_id": 1,
                "title": "Sample",
                "slug": "sample—text",
                "post_stream": {
                    "posts": [
                        {
                            "id": 11,
                            "cooked": (
                                "<a href='/u/evilnick'>@evilnick</a>"
                                "<a>No Link</a>"
                                "<a></a>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ],
                },
            }
        )

        self.assertIn(
            '<a href="https://base.url/u/evilnick">@evilnick</a>',
            parsed_topic["body_html"],
        )

    def test_parser_upload(self):
        discourse_api = DiscourseAPI("https://base.url", session=MagicMock())

        parser = BaseParser(
            api=discourse_api,
            index_topic_id=1,
            url_prefix="/",
        )

        parsed_topic = parser.parse_topic(
            {
                "id": 1,
                "category_id": 1,
                "title": "Sample",
                "slug": "sample—text",
                "post_stream": {
                    "posts": [
                        {
                            "id": 11,
                            "cooked": (
                                "<a href='/uploads/test.png'>"
                                "<img src='test.png' srcset='test.png' />"
                                "</a>"
                                "<a href='/uploads/test2.png'>"
                                "<img src='test2.png' />"
                                "</a>"
                                "<a>No Link</a>"
                                "<a></a>"
                            ),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ],
                },
            }
        )

        self.assertIn(
            (
                '<a href="https://base.url/uploads/test.png">'
                '<img src="test.png"/>'
                "</a>"
            ),
            parsed_topic["body_html"],
        )
        self.assertIn(
            (
                '<a href="https://base.url/uploads/test2.png">'
                '<img src="test2.png"/>'
                "</a>"
            ),
            parsed_topic["body_html"],
        )

    def test_emdash_in_slug(self):
        discourse_api = DiscourseAPI("https://base.url", session=MagicMock())

        parser = BaseParser(
            api=discourse_api,
            index_topic_id=1,
            url_prefix="/",
        )

        parsed_topic = parser.parse_topic(
            {
                "id": 1,
                "category_id": 1,
                "title": "Sample",
                "slug": "sample—text",
                "post_stream": {
                    "posts": [
                        {
                            "id": 11,
                            "cooked": ("empty"),
                            "updated_at": "2018-10-02T12:45:44.259Z",
                        }
                    ],
                },
            }
        )

        self.assertEqual("/t/sample--text/1", parsed_topic["topic_path"])


class TestDocParser(unittest.TestCase):
    def setUp(self):
        # Suppress annoying warnings from HTTPretty
        # See: https://github.com/gabrielfalcao/HTTPretty/issues/368
        warnings.filterwarnings(
            "ignore", category=ResourceWarning, message="unclosed.*"
        )

        # Enable HTTPretty and set up mock URLs
        httpretty.enable()
        self.addCleanup(httpretty.disable)
        self.addCleanup(httpretty.reset)
        # Index page with navigation, URL map and redirects
        httpretty.register_uri(
            httpretty.GET,
            "https://discourse.example.com/t/34.json",
            body=json.dumps(
                {
                    "id": 34,
                    "category_id": 2,
                    "title": "An index page",
                    "slug": "an-index-page",
                    "post_stream": {
                        "posts": [
                            {
                                "id": 3434,
                                "cooked": EXAMPLE_CONTENT,
                                "updated_at": "2018-10-02T12:45:44.259Z",
                            }
                        ]
                    },
                }
            ),
            content_type="application/json",
        )

        discourse_api = DiscourseAPI(
            base_url="https://discourse.example.com/",
            session=requests.Session(),
        )

        self.parser = DocParser(
            api=discourse_api,
            index_topic_id=34,
            url_prefix="/",
        )
        self.parser.parse()
        self.index = self.parser.parse_topic(self.parser.index_topic)

    def test_index_has_no_nav(self):
        soup = BeautifulSoup(self.index["body_html"], features="lxml")

        # Check body
        self.assertEqual(soup.p.string, "Some homepage content")

        # Check navigation
        self.assertIsNone(soup.h1)

        # Check URL map worked
        self.assertIsNone(soup.details)
        self.assertNotIn(
            '<a href="/t/page-a/10">Page A</a>',
            soup.decode_contents(),
        )

    def test_get_section(self):
        soup = BeautifulSoup(
            self.parser.index_topic["post_stream"]["posts"][0]["cooked"],
            features="lxml",
        )
        section = self.parser._get_section(soup, "Navigation")
        self.assertEqual(len(section("table")), 1)
        self.assertEqual(len(section.table("tr")), 3)
        last_entry = section.table("tr")[-1]
        self.assertEqual(
            list(last_entry.stripped_strings), ["1", "/page-z", "Page Z"]
        )

    def test_get_sections(self):
        soup = BeautifulSoup(
            self.parser.index_topic["post_stream"]["posts"][0]["cooked"],
            features="lxml",
        )
        soup_str = str(soup)
        sections = self.parser._get_sections(soup)
        self.assertEqual(len(sections), 2)
        first = sections[0]
        self.assertEqual(first.keys(), {"title", "content", "slug"})
        self.assertEqual(first["title"], "Navigation")
        self.assertEqual(first["slug"], "navigation")
        self.assertEqual(first["content"][:9], "<details>")
        self.assertEqual(first["content"][-10:], "</details>")
        self.assertEqual(len(first["content"]), 353)
        second = sections[1]
        self.assertEqual(second.keys(), {"title", "content", "slug"})
        self.assertEqual(second["title"], "Redirects")
        self.assertEqual(second["slug"], "redirects")
        self.assertEqual(second["content"][:9], "<details>")
        self.assertEqual(second["content"][-10:], "</details>")
        self.assertEqual(len(second["content"]), 286)
        self.assertEqual(str(soup), soup_str)

    def test_nav(self):
        navigation = self.parser.navigation
        page_a = navigation["nav_items"][0]
        self.assertEqual(page_a["path"], "/a")
        self.assertEqual(page_a["navlink_text"], "Page A")
        page_z = page_a["children"][0]
        self.assertEqual(page_z["path"], "/page-z")
        self.assertEqual(page_z["navlink_text"], "Page Z")

    def test_active_topic(self):
        for page_id in [10, 26]:
            httpretty.register_uri(
                httpretty.GET,
                f"https://discourse.example.com/t/{page_id}.json",
                body=json.dumps(
                    {
                        "id": page_id,
                        "category_id": 2,
                        "title": "A topic page",
                        "slug": "a-page",
                        "post_stream": {
                            "posts": [
                                {
                                    "id": 3434,
                                    "cooked": "<h1>Content</h1>",
                                    "updated_at": "2023-04-01T12:34:56.789Z",
                                }
                            ]
                        },
                    }
                ),
                content_type="application/json",
            )
        self.assertEqual(self.parser.active_topic_id, 34)
        root_navigation = self.parser.navigation

        root_page_a = root_navigation["nav_items"][0]
        self.assertFalse(root_page_a["is_active"])
        self.assertFalse(root_page_a["has_active_child"])

        # Simulate clicking on child page
        self.parser.parse_topic(self.parser.api.get_topic(10))
        self.assertEqual(self.parser.active_topic_id, 10)
        child = self.parser.navigation["nav_items"][0]
        self.assertTrue(child["is_active"])
        self.assertFalse(child["has_active_child"])

        # Simulate clicking on grand-child page
        self.parser.parse_topic(self.parser.api.get_topic(26))
        self.assertEqual(self.parser.active_topic_id, 26)
        child = self.parser.navigation["nav_items"][0]
        self.assertFalse(child["is_active"])
        self.assertTrue(child["has_active_child"])
        grandchild = child["children"][0]
        self.assertTrue(grandchild["is_active"])
        self.assertFalse(grandchild["has_active_child"])

        # Simulate clicking on root
        self.parser.parse_topic(self.parser.index_topic)
        child = self.parser.navigation["nav_items"][0]
        self.assertFalse(child["is_active"])
        self.assertFalse(child["has_active_child"])

    def test_versions(self):
        versions = self.parser.versions
        self.assertEqual(len(versions), 1)
        version = versions[0]
        self.assertEqual(version["index"], int(self.parser.index_topic_id))
        self.assertEqual(version["path"], "")
        self.assertEqual(version["version"], "latest")
        self.assertNotEqual(version["nav_items"], [])

    def test_redirect_map(self):
        self.assertEqual(
            self.parser.redirect_map,
            {"/redir-a": "/a", "/example/page": "https://example.com/page"},
        )

        self.assertEqual(self.parser.warnings, [])

    def test_url_map(self):
        self.assertEqual(
            self.parser.url_map,
            {
                10: "/a",
                26: "/page-z",
                34: "/",
                "/": 34,
                "/a": 10,
                "/page-z": 26,
            },
        )

        self.assertEqual(self.parser.warnings, [])



from canonicalwebteam.discourse.app import (  # noqa
    Docs,  # noqa
    EngagePages,  # noqa
    Tutorials,  # noqa
)
from canonicalwebteam.discourse.models import DiscourseAPI  # noqa
from canonicalwebteam.discourse.parsers import (  # noqa
    DocParser,  # noqa
    TutorialParser,  # noqa
)



import os
import requests

import flask
from vcr_unittest import VCRTestCase

from canonicalwebteam.discourse import DiscourseAPI, EngagePages

this_dir = os.path.dirname(os.path.realpath(__file__))


class TestDiscourseAPI(VCRTestCase):
    def _get_vcr_kwargs(self):
        """
        This removes the authorization header
        from VCR so we don't record auth parameters
        """
        return {"filter_headers": ["Authorization"]}

    def setUp(self):
        app = flask.Flask("test-app")
        app.url_map.strict_slashes = False
        app.template_folder = f"{this_dir}/fixtures/templates"
        app.testing = True
        session = requests.Session()

        self.discourse_api = DiscourseAPI(
            base_url="https://discourse.ubuntu.com/",
            session=session,
        )
        self.engage_pages = EngagePages(
            category_id=51,
            api=self.discourse_api,
            page_type="engage-pages",
        )

        self.takeovers = EngagePages(
            category_id=106,
            api=self.discourse_api,
            page_type="takeovers",
        )

        self.client = app.test_client()
        return super().setUp()

    def test_get_topic(self):
        response = self.discourse_api.get_topic(17275)

        self.assertEqual(response["id"], 17275)



# Standard library
import os
import unittest
import warnings

# Packages
import flask
import httpretty
import requests

# Local
from canonicalwebteam.discourse import DiscourseAPI, Tutorials, TutorialParser
from tests.fixtures.forum_mock import register_uris


this_dir = os.path.dirname(os.path.realpath(__file__))
template_folder = f"{this_dir}/fixtures/templates"


# XXX: This has a lot of setup and no tests, none of the setup will be
# run without any tests to run it for...
class TestApp(unittest.TestCase):
    def setUp(self):
        """
        Set up Flask app with Discourse extension for testing
        And set up mocking for discourse.example.com
        """

        # Suppress annoying warnings from HTTPretty
        # See: https://github.com/gabrielfalcao/HTTPretty/issues/368
        warnings.filterwarnings(
            "ignore", category=ResourceWarning, message="unclosed.*"
        )

        # Enable HTTPretty and set up mock URLs
        httpretty.enable()
        register_uris()

        app = flask.Flask("main", template_folder=template_folder)
        app_no_nav = flask.Flask("no-nav", template_folder=template_folder)
        app_no_mappings = flask.Flask(
            "no-mappings", template_folder=template_folder
        )
        app_broken_mappings = flask.Flask(
            "broken-mappings", template_folder=template_folder
        )
        app_no_category = flask.Flask(
            "no-category", template_folder=template_folder
        )
        app_url_prefix = flask.Flask(
            "url-prefix", template_folder=template_folder
        )

        app.testing = True
        app_no_nav.testing = True
        app_no_mappings.testing = True
        app_broken_mappings.testing = True
        app_no_category.testing = True
        app_url_prefix.testing = True

        discourse_api = DiscourseAPI(
            base_url="https://discourse.example.com/",
            session=requests.Session(),
        )

        Tutorials(
            parser=TutorialParser(
                api=discourse_api,
                index_topic_id=34,
                url_prefix="/",
            ),
            document_template="document.html",
            url_prefix="/",
        ).init_app(app)

        Tutorials(
            parser=TutorialParser(
                api=discourse_api,
                index_topic_id=42,
                url_prefix="/",
            ),
            document_template="document.html",
            url_prefix="/",
        ).init_app(app_no_nav)

        Tutorials(
            parser=TutorialParser(
                api=discourse_api,
                index_topic_id=35,
                url_prefix="/",
            ),
            document_template="document.html",
            url_prefix="/",
        ).init_app(app_no_mappings)

        Tutorials(
            parser=TutorialParser(
                api=discourse_api,
                index_topic_id=36,
                url_prefix="/",
            ),
            document_template="document.html",
            url_prefix="/",
        ).init_app(app_broken_mappings)

        Tutorials(
            parser=TutorialParser(
                api=discourse_api, index_topic_id=37, url_prefix="/"
            ),
            document_template="document.html",
            url_prefix="/",
        ).init_app(app_no_category)

        Tutorials(
            parser=TutorialParser(
                api=discourse_api, index_topic_id=38, url_prefix="/docs"
            ),
            document_template="document.html",
            url_prefix="/tutorials",
        ).init_app(app_url_prefix)

        self.client = app.test_client()
        self.client_no_nav = app_no_nav.test_client()
        self.client_no_mappings = app_no_mappings.test_client()
        self.client_broken_mappings = app_broken_mappings.test_client()
        self.client_no_category = app_no_category.test_client()
        self.client_url_prefix = app_url_prefix.test_client()

    def tearDown(self):
        httpretty.disable()
        httpretty.reset()



#! /usr/bin/env python3

from setuptools import setup, find_packages

setup(
    name="canonicalwebteam.discourse",
    version="5.6.0",
    author="Canonical webteam",
    author_email="webteam@canonical.com",
    url="https://github.com/canonical/canonicalwebteam.discourse",
    description=(
        "Flask extension to integrate discourse content generated "
        "to docs to your website."
    ),
    packages=find_packages(),
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    install_requires=[
        "Flask>=1.0.2",
        "beautifulsoup4",
        "humanize",
        "lxml",
        "requests",
        "python-dateutil",
        "validators",
    ],
)
